@@ -33,6 +33,7 @@
 #endif
 
 #include "mod_mono.h"
+#include "mono-io-portability.h"
 
 DEFINE_MODULE (mono_module);
 
@@ -63,10 +64,26 @@ typedef struct {
 	uint32_t handled_requests;
 	time_t start_time;
 	char restart_issued;
+	char starting;
 	int active_requests;
 	int waiting_requests;
+	int accepting_requests;
 } dashboard_data;
 
+#define INITIAL_DATA_MAX_ALLOCA_SIZE 8192
+typedef struct {
+	size_t      method_len;
+	size_t      server_hostname_len;
+	size_t      uri_len;
+	size_t      args_len;
+	size_t      protocol_len;
+	size_t      local_ip_len;
+	size_t      remote_ip_len;
+	const char *remote_name;
+	size_t      remote_name_len;
+	size_t      filename_len;
+} initial_data_info;
+
 typedef struct xsp_data {
 	char is_default;
 	char *alias;
@@ -87,6 +104,7 @@ typedef struct xsp_data {
 	char *max_memory;
 	char *debug;
 	char *env_vars;
+	char *iomap;
 	char status; /* One of the FORK_* in the enum above.
 		      * Don't care if run_xsp is "false" */
 	char is_virtual; /* is the server virtual? */
@@ -102,7 +120,8 @@ typedef struct xsp_data {
 
 	/* other settings */
 	unsigned char no_flush;
-	
+	int portability_level;
+
 #ifndef APACHE13
 	apr_shm_t *dashboard_shm;
 	dashboard_data *dashboard;
@@ -341,7 +360,7 @@ apache_get_groupid ()
 #endif
 }
 
-inline static char *
+inline static const char *
 apache_get_username ()
 {
 #ifdef HAVE_UNIXD
@@ -357,6 +376,7 @@ ensure_dashboard_initialized (module_cfg *config, xsp_data *xsp, apr_pool_t *p)
 {
 	apr_status_t rv;
 	mode_t old_umask;
+	int is_global;
 #if defined (APR_HAS_USER)
 	apr_uid_t cur_uid;
 	apr_gid_t cur_gid;
@@ -371,6 +391,11 @@ ensure_dashboard_initialized (module_cfg *config, xsp_data *xsp, apr_pool_t *p)
 	}
 #endif
 
+	if (strcmp ("XXGLOBAL", xsp->alias))
+		is_global = 1;
+	else
+		is_global = 0;
+
 #if defined (APR_HAS_USER)
 	if (apr_uid_current (&cur_uid, &cur_gid, p) == APR_SUCCESS && cur_uid == 0) {
 		DEBUG_PRINT (2, "Temporarily switching to target uid/gid");
@@ -422,9 +447,10 @@ ensure_dashboard_initialized (module_cfg *config, xsp_data *xsp, apr_pool_t *p)
 		} else {
 			DEBUG_PRINT (1, "removing dashboard file '%s'", xsp->dashboard_file);
 			if (unlink (xsp->dashboard_file) == -1 && errno != ENOENT) {
-				ap_log_error (APLOG_MARK, APLOG_CRIT, STATCODE_AND_SERVER (rv),
-				      "Failed to attach to existing dashboard, and removing dashboard file '%s' failed (%s). Further action impossible.",
-				      xsp->dashboard_file, strerror (errno));
+				if (!is_global)
+					ap_log_error (APLOG_MARK, APLOG_CRIT, STATCODE_AND_SERVER (rv),
+						      "Failed to attach to existing dashboard, and removing dashboard file '%s' failed (%s). Further action impossible.",
+						      xsp->dashboard_file, strerror (errno));
 				goto restore_creds;
 			}
 	
@@ -452,6 +478,8 @@ ensure_dashboard_initialized (module_cfg *config, xsp_data *xsp, apr_pool_t *p)
 				xsp->dashboard->restart_issued = 0;
 				xsp->dashboard->active_requests = 0;
 				xsp->dashboard->waiting_requests = 0;
+				xsp->dashboard->starting = 0;
+				xsp->dashboard->accepting_requests = 1;
 			}
 		}
 	}
@@ -511,6 +539,8 @@ add_xsp_server (apr_pool_t *pool, const char *alias, module_cfg *config, int is_
 	server->max_memory = NULL;
 	server->debug = NULL;
 	server->env_vars = NULL;
+	server->iomap = NULL;
+	server->portability_level = PORTABILITY_UNKNOWN;
 	server->status = FORK_NONE;
 	server->is_virtual = is_virtual;
 	server->start_attempts = "3";
@@ -719,7 +749,6 @@ create_mono_server_config (apr_pool_t *p, server_rec *s)
 	server->auto_app_set = FALSE;
 	
 	add_xsp_server (p, "XXGLOBAL", server, FALSE, FALSE);
-	server->servers [0].filename = get_default_global_socket_name (p, SOCKET_FILE);
 	return server;
 }
 
@@ -885,8 +914,11 @@ read_data_string (apr_pool_t *pool, apr_socket_t *sock, char **ptr, int32_t *siz
 }
 
 static int
-send_entire_file (request_rec *r, const char *filename, int *result)
+send_entire_file (request_rec *r, const char *filename, int *result, xsp_data *xsp)
 {
+	int retval = 0;
+	gchar *file_path = mono_portability_find_file (xsp->portability_level, filename, TRUE);
+
 #ifdef APACHE2
 #  ifdef APR_LARGEFILE 
 #      define MODMONO_LARGE APR_LARGEFILE
@@ -899,18 +931,20 @@ send_entire_file (request_rec *r, const char *filename, int *result)
 	apr_size_t nbytes;
 	const apr_int32_t flags = APR_READ | APR_SENDFILE_ENABLED | MODMONO_LARGE;
 
-	st = apr_file_open (&file, filename, flags, APR_OS_DEFAULT, r->pool);
+	st = apr_file_open (&file, file_path ? file_path : filename, flags, APR_OS_DEFAULT, r->pool);
 	if (st != APR_SUCCESS) {
-		DEBUG_PRINT (1, "file_open FAILED");
+		DEBUG_PRINT (1, "file_open FAILED (path: %s)", file_path ? file_path : filename);
 		*result = HTTP_FORBIDDEN; 
-		return -1;
+		retval = -1;
+		goto finish;
 	}
 
 	st = apr_file_info_get (&info, APR_FINFO_SIZE, file);
 	if (st != APR_SUCCESS) {
 		DEBUG_PRINT (1, "info_get FAILED");
 		*result = HTTP_FORBIDDEN; 
-		return -1;
+		retval = -1;
+		goto finish;
 	}
 
 	st = ap_send_fd (file, r, 0, info.size, &nbytes);
@@ -918,29 +952,33 @@ send_entire_file (request_rec *r, const char *filename, int *result)
 	if (nbytes < 0) {
 		DEBUG_PRINT (1, "SEND FAILED");
 		*result = HTTP_INTERNAL_SERVER_ERROR;
-		return -1;
+		retval = -1;
+		goto finish;
 	}
-
-	return 0;
 #else
 	FILE *fp;
 
-	fp = fopen (filename, "rb");
+	fp = fopen (file_path ? file_path : filename, "rb");
 	if (fp == NULL) {
-		DEBUG_PRINT (1, "file_open FAILED");
+		DEBUG_PRINT (1, "file_open FAILED (path: %s)", file_path ? file_path : filename);
 		*result = HTTP_FORBIDDEN; 
-		return -1;
+		retval = -1;
+		goto finish;
 	}
 
 	if (ap_send_fd (fp, r) < 0) {
 		fclose (fp);
 		*result = HTTP_INTERNAL_SERVER_ERROR;
-		return -1;
+		retval = -1;
+		goto finish;
 	}
-		
+
 	fclose (fp);
-	return 0;
 #endif
+  finish:
+	if (file_path)
+		g_free (file_path);
+	return retval;
 }
 
 static int
@@ -1111,6 +1149,7 @@ do_command (int command, apr_socket_t *sock, request_rec *r, int *result, xsp_da
 				abort();
 			}
 			str = get_client_block_buffer (r, (uint32_t) i, &actual_size);
+			DEBUG_PRINT (2, "requested size == %u, actual size == %u", i, actual_size);
 			i = ap_get_client_block (r, str, actual_size);
 			i = LE_FROM_INT (i);
 			status = write_data (sock, &i, sizeof (int32_t));
@@ -1163,7 +1202,7 @@ do_command (int command, apr_socket_t *sock, request_rec *r, int *result, xsp_da
 				status = -1;
 				break;
 			}
-			status = send_entire_file (r, str, result);
+			status = send_entire_file (r, str, result, xsp);
 			if (status == -1)
 				error_message = "failed to send file (file data)";
 			break;
@@ -1299,7 +1338,7 @@ get_default_socket_name (apr_pool_t *pool, const char *alias, const char *base)
 }
 
 static apr_status_t 
-try_connect (xsp_data *conf, apr_socket_t **sock, apr_pool_t *pool)
+try_connect (xsp_data *conf, apr_socket_t **sock, apr_int32_t family, apr_pool_t *pool)
 {
 	char *error;
 	struct sockaddr_un unix_address;
@@ -1328,7 +1367,7 @@ try_connect (xsp_data *conf, apr_socket_t **sock, apr_pool_t *pool)
 		apr_sockaddr_t *sa;
 
 		la = conf->listen_address ? conf->listen_address : LISTEN_ADDRESS;
-		rv = apr_sockaddr_info_get (&sa, la, APR_INET,
+		rv = apr_sockaddr_info_get (&sa, la, family,
 					    atoi (conf->listen_port), 0, pool);
 
 		if (rv != APR_SUCCESS) {
@@ -1479,6 +1518,19 @@ set_process_limits (int max_cpu_time, int max_memory)
 	set_process_limits2 (RLIMIT_DATA, max_memory, "memory (data segment)");
 }
 
+static void
+set_null_stdout ()
+{
+#ifndef WIN32
+       int fd;
+
+       fd = open ("/dev/null", O_WRONLY);
+       if (fd >= 0) {
+               dup2 (fd, 1);
+       }
+#endif
+}
+
 static void
 fork_mod_mono_server (apr_pool_t *pool, xsp_data *config)
 {
@@ -1495,6 +1547,7 @@ fork_mod_mono_server (apr_pool_t *pool, xsp_data *config)
 	int max_cpu_time = 0;
 	int status;
 	char is_master;
+	sigset_t sigset;
 #if defined (APR_HAS_USER)
 	apr_uid_t cur_uid;
 	apr_gid_t cur_gid;
@@ -1507,7 +1560,10 @@ fork_mod_mono_server (apr_pool_t *pool, xsp_data *config)
 		return;
 	}
 #endif
-  
+	is_master = (0 == strcmp ("XXGLOBAL", config->alias));
+	if (is_master && config->listen_port == NULL && config->filename == NULL)
+		config->filename = get_default_global_socket_name (pool, SOCKET_FILE);
+
 	/* Running mod-mono-server not requested */
 	if (!strcasecmp (config->run_xsp, "false")) {
 		DEBUG_PRINT (1, "Not running mod-mono-server: %s", config->run_xsp);
@@ -1516,7 +1572,6 @@ fork_mod_mono_server (apr_pool_t *pool, xsp_data *config)
 		return;
 	}
 
-	is_master = (0 == strcmp ("XXGLOBAL", config->alias));
 	/*
 	 * At least one of MonoApplications, MonoApplicationsConfigFile or
 	 * MonoApplicationsConfigDir must be specified, except for the 'global'
@@ -1560,6 +1615,9 @@ fork_mod_mono_server (apr_pool_t *pool, xsp_data *config)
 
 	set_environment_variables (pool, config->env_vars);
 
+	if (config->iomap && *config->iomap)
+		SETENV (pool, "MONO_IOMAP", config->iomap);
+
 	pid = fork ();
 	if (pid > 0) {
 		wait (&status);
@@ -1572,7 +1630,7 @@ fork_mod_mono_server (apr_pool_t *pool, xsp_data *config)
 		exit (0);
 
 	setsid ();
-	chdir ("/");
+	status = chdir ("/");
 
 #if defined (APR_HAS_USER)
 	/*
@@ -1609,6 +1667,8 @@ fork_mod_mono_server (apr_pool_t *pool, xsp_data *config)
 
 #ifdef DEBUG
 	dup2 (2, 1);
+#else
+	set_null_stdout ();
 #endif
 	for (i = getdtablesize () - 1; i >= 3; i--)
 		close (i);
@@ -1641,6 +1701,14 @@ fork_mod_mono_server (apr_pool_t *pool, xsp_data *config)
 	if (config->debug && !strcasecmp (config->debug, "True"))
 		SETENV (pool, "MONO_OPTIONS", "--debug");
 
+#ifdef REMOVE_DISPLAY
+#warning mod_mono is compiled with support for removing the DISPLAY variable (Bug 464225)
+#ifdef HAVE_UNSETENV
+	unsetenv ("DISPLAY");
+#else
+#warning unsetenv (2) is not available!
+#endif
+#endif
 	memset (argv, 0, sizeof (char *) * MAXARGS);
 	argi = 0;
 	argv [argi++] = config->server_path;
@@ -1702,6 +1770,14 @@ fork_mod_mono_server (apr_pool_t *pool, xsp_data *config)
 		      argv [5], argv [6], argv [7], argv [8], 
 		      argv [9], argv [10], argv [11], argv [12]);
 
+	/* Unblock signals Mono uses: see bug #472732 */
+	sigemptyset (&sigset);
+	sigaddset (&sigset, SIGPWR);
+	sigaddset (&sigset, SIGXCPU);
+	sigaddset (&sigset, 33);
+	sigaddset (&sigset, 35);
+	sigprocmask (SIG_UNBLOCK, &sigset, NULL);
+
 	execv (argv [0], argv);
 	ap_log_error (APLOG_MARK, APLOG_ERR, STATUS_AND_SERVER,
 		      "Failed running '%s %s %s %s %s %s %s %s %s %s %s %s %s'. Reason: %s",
@@ -1718,7 +1794,33 @@ setup_socket (apr_socket_t **sock, xsp_data *conf, apr_pool_t *pool)
 	apr_status_t rv;
 	int family, proto;
 
-	family = (conf->listen_port != NULL) ? AF_UNSPEC : PF_UNIX;
+	if (conf->listen_port != NULL) {
+#if APR_HAVE_IPV6
+		apr_status_t rv;
+		apr_sockaddr_t *sa;
+		char *la = NULL;
+#endif
+		family = AF_UNSPEC;
+
+#if APR_HAVE_IPV6
+		la = conf->listen_address ? conf->listen_address : LISTEN_ADDRESS;
+		rv = apr_sockaddr_info_get (&sa, la, family,
+					    atoi (conf->listen_port), 
+					    APR_IPV6_ADDR_OK,
+					    pool);
+
+		if (rv != APR_SUCCESS) {
+			ap_log_error (APLOG_MARK, APLOG_ERR, STATUS_AND_SERVER,
+				      "mod_mono: error in address ('%s') or port ('%s'). Assuming AF_UNSPEC.",
+				      la, conf->listen_port);
+		} else {
+			family = sa->family;
+		}
+
+#endif
+	} else {
+		family = PF_UNIX;
+	}
 	/* APR_PROTO_TCP = 6 */
 	proto = (family == AF_UNSPEC) ? 6 : 0;
 #ifdef APACHE2
@@ -1735,27 +1837,33 @@ setup_socket (apr_socket_t **sock, xsp_data *conf, apr_pool_t *pool)
 		return rv;
 	}
 
-	rv = try_connect (conf, sock, pool);
+	rv = try_connect (conf, sock, family, pool);
 	DEBUG_PRINT (1, "try_connect: %d", (int) rv);
 	return rv;
 }
 
 static int
-write_string_to_buffer (char *buffer, int offset, const char *str)
+write_string_to_buffer (char *buffer, int offset, const char *str, size_t str_length)
 {
-	int tmp;
-	int le;
-
+	uint32_t le;
+	uint32_t tmp;
+	
 	buffer += offset;
-	tmp = (str != NULL) ? strlen (str) : 0;
+	if (str && !str_length) {
+		tmp = strlen (str);
+		le = LE_FROM_INT (tmp);
+	} else
+		tmp = (uint32_t)str_length;
+	
 	le = LE_FROM_INT (tmp);
-	memcpy (buffer, &le, sizeof (int32_t));
+	
+	memcpy (buffer, &le, sizeof (uint32_t));
 	if (tmp > 0) {
-		buffer += sizeof (int32_t);
+		buffer += sizeof (uint32_t);
 		memcpy (buffer, str, tmp);
 	}
 
-	return tmp + sizeof (int32_t);
+	return tmp + sizeof (uint32_t);
 }
 
 static int32_t
@@ -1770,7 +1878,7 @@ get_table_send_size (apr_table_t *table)
 	if (elts->nelts == 0)
 		return sizeof (int32_t);
 
-	size = sizeof (int32_t);
+	size = sizeof (int32_t) * 2;
 	t_elt = (const apr_table_entry_t *) (elts->elts);
 	t_end = t_elt + elts->nelts;
 
@@ -1793,8 +1901,10 @@ write_table_to_buffer (char *buffer, apr_table_t *table)
 	const apr_table_entry_t *t_elt;
 	const apr_table_entry_t *t_end;
 	char *ptr;
-	int32_t count = 0;
-
+	int32_t count = 0, size;
+	char *countLocation = buffer + sizeof (int32_t);
+	char *sizeLocation = buffer;
+	
 	elts = apr_table_elts (table);
 	if (elts->nelts == 0) { /* size is sizeof (int32_t) */
 		int32_t *i32 = (int32_t *) buffer;
@@ -1803,16 +1913,17 @@ write_table_to_buffer (char *buffer, apr_table_t *table)
 	}
 
 	ptr = buffer;
+	/* the size is set after the loop */
 	/* the count is set after the loop */
-	ptr += sizeof (int32_t);
+	ptr += sizeof (int32_t) * 2;
 	t_elt = (const apr_table_entry_t *) (elts->elts);
 	t_end = t_elt + elts->nelts;
-
+	
 	do {
 		if (t_elt->val != NULL) {
 			DEBUG_PRINT (3, "%s: %s", t_elt->key, t_elt->val);
-			ptr += write_string_to_buffer (ptr, 0, t_elt->key);
-			ptr += write_string_to_buffer (ptr, 0, t_elt->val);
+			ptr += write_string_to_buffer (ptr, 0, t_elt->key, 0);
+			ptr += write_string_to_buffer (ptr, 0, t_elt->val, 0);
 			count++;
 		}
 
@@ -1820,7 +1931,11 @@ write_table_to_buffer (char *buffer, apr_table_t *table)
 	} while (t_elt < t_end);
 
 	count = LE_FROM_INT (count);
-	memcpy (buffer, &count, sizeof (int32_t));
+	memcpy (countLocation, &count, sizeof (int32_t));
+	size = (ptr - buffer) - sizeof (int32_t);
+	size = LE_FROM_INT (size);
+	memcpy (sizeLocation, &size, sizeof (int32_t));
+	
 	return (ptr - buffer);
 }
 
@@ -1839,62 +1954,97 @@ send_table (apr_pool_t *pool, apr_table_t *table, apr_socket_t *sock)
 static int
 send_initial_data (request_rec *r, apr_socket_t *sock, char auto_app)
 {
-	int i;
-	char *str, *ptr;
-	int size;
-	server_rec *s = r->server;
+	uint32_t           i;
+	char              *str, *ptr;
+	uint32_t           size;
+	server_rec        *s = r->server;
+	initial_data_info  info;
+	
+	DEBUG_PRINT (2, "Send init");
+	size = 1 + sizeof (size);
 
-	DEBUG_PRINT (2, "Send init1");
-	size = 1;
-	size += ((r->method != NULL) ? strlen (r->method) : 0) + sizeof (int32_t);
-	if (s != NULL)
-		size += ((s->is_virtual && s->server_hostname != NULL) ? strlen (s->server_hostname) : 0) + sizeof (int32_t);
-	else
+	info.method_len = ((r->method != NULL) ? strlen (r->method) : 0);
+	size += info.method_len + sizeof (int32_t);
+	if (s != NULL) {
+		info.server_hostname_len = ((s->is_virtual && s->server_hostname != NULL) ? strlen (s->server_hostname) : 0);
+		size += info.server_hostname_len + sizeof (int32_t);
+	} else {
+		info.server_hostname_len = 0;
 		size += sizeof (int32_t);
-	size += ((r->uri != NULL) ? strlen (r->uri) : 0) + sizeof (int32_t);
-	size += ((r->args != NULL) ? strlen (r->args) : 0) + sizeof (int32_t);
-	size += ((r->protocol != NULL) ? strlen (r->protocol) : 0) + sizeof (int32_t);
-	size += strlen (r->connection->local_ip) + sizeof (int32_t);
+	}
+
+	info.uri_len = ((r->uri != NULL) ? strlen (r->uri) : 0);
+	size += info.uri_len + sizeof (int32_t);
+
+	info.args_len = ((r->args != NULL) ? strlen (r->args) : 0);
+	size += info.args_len + sizeof (int32_t);
+
+	info.protocol_len = ((r->protocol != NULL) ? strlen (r->protocol) : 0);
+	size += info.protocol_len + sizeof (int32_t);
+
+	info.local_ip_len = strlen (r->connection->local_ip);
+	size += info.local_ip_len + sizeof (int32_t);
+	
 	size += sizeof (int32_t);
-	size += strlen (r->connection->remote_ip) + sizeof (int32_t);
+
+	info.remote_ip_len = strlen (r->connection->remote_ip);
+	size += info.remote_ip_len + sizeof (int32_t);
+	
 	size += sizeof (int32_t);
-	size += strlen (connection_get_remote_name (r)) + sizeof (int32_t);
+
+	info.remote_name = connection_get_remote_name (r);
+	info.remote_name_len = strlen (info.remote_name);
+	
+	size += info.remote_name_len + sizeof (int32_t);
+	
 	size += get_table_send_size (r->headers_in);
+	
 	size++; /* byte. TRUE->auto_app, FALSE: configured application */
 	if (auto_app != FALSE) {
 		if (r->filename != NULL) {
-			size += strlen (r->filename) + sizeof (int32_t);
+			info.filename_len = strlen (r->filename);
+			size += info.filename_len + sizeof (int32_t);
 		} else {
+			info.filename_len = 0;
 			auto_app = FALSE;
 		}
-	}
+	} else
+		info.filename_len = 0;
 
-	ptr = str = apr_pcalloc (r->pool, size);
-	*ptr++ = PROTOCOL_VERSION; /* version. Keep in sync with ModMonoRequest. */
-	ptr += write_string_to_buffer (ptr, 0, r->method);
-	if (s != NULL)
-		ptr += write_string_to_buffer (ptr, 0, (s->is_virtual ? s->server_hostname : NULL));
+	DEBUG_PRINT (2, "Initial data size: %u", size);
+
+	if (size <= INITIAL_DATA_MAX_ALLOCA_SIZE)
+		ptr = str = alloca (size);
 	else
-		ptr += write_string_to_buffer (ptr, 0, NULL);
-	ptr += write_string_to_buffer (ptr, 0, r->uri);
-	ptr += write_string_to_buffer (ptr, 0, r->args);
-	ptr += write_string_to_buffer (ptr, 0, r->protocol);
+		ptr = str = apr_pcalloc (r->pool, size);
+	*ptr++ = (char)PROTOCOL_VERSION; /* version. Keep in sync with ModMonoRequest. */
+	i = LE_FROM_INT (size);
+	memcpy (ptr, &i, sizeof (i));
+	ptr += sizeof (int32_t);
+	ptr += write_string_to_buffer (ptr, 0, r->method, info.method_len);
+	if (s != NULL)
+		ptr += write_string_to_buffer (ptr, 0, (s->is_virtual ? s->server_hostname : NULL), info.server_hostname_len);
+	else
+		ptr += write_string_to_buffer (ptr, 0, NULL, 0);
+	ptr += write_string_to_buffer (ptr, 0, r->uri, info.uri_len);
+	ptr += write_string_to_buffer (ptr, 0, r->args, info.args_len);
+	ptr += write_string_to_buffer (ptr, 0, r->protocol, info.protocol_len);
 
-	ptr += write_string_to_buffer (ptr, 0, r->connection->local_ip);
+	ptr += write_string_to_buffer (ptr, 0, r->connection->local_ip, info.local_ip_len);
 	i = request_get_server_port (r);
 	i = LE_FROM_INT (i);
-	memcpy (ptr, &i, sizeof (int32_t));
+	memcpy (ptr, &i, sizeof (i));
 	ptr += sizeof (int32_t);
-	ptr += write_string_to_buffer (ptr, 0, r->connection->remote_ip);
+	ptr += write_string_to_buffer (ptr, 0, r->connection->remote_ip, info.remote_ip_len);
 	i = connection_get_remote_port (r->connection);
 	i = LE_FROM_INT (i);
-	memcpy (ptr, &i, sizeof (int32_t));
+	memcpy (ptr, &i, sizeof (i));
 	ptr += sizeof (int32_t);
-	ptr += write_string_to_buffer (ptr, 0, connection_get_remote_name (r));
+	ptr += write_string_to_buffer (ptr, 0, info.remote_name, info.remote_name_len);
 	ptr += write_table_to_buffer (ptr, r->headers_in);
 	*ptr++ = auto_app;
 	if (auto_app != FALSE)
-		ptr += write_string_to_buffer (ptr, 0, r->filename);
+		ptr += write_string_to_buffer (ptr, 0, r->filename, info.filename_len);
 
 	if (write_data (sock, str, size) != size)
 		return -1;
@@ -1905,6 +2055,14 @@ send_initial_data (request_rec *r, apr_socket_t *sock, char auto_app)
 static int
 increment_active_requests (xsp_data *conf)
 {
+#ifndef APACHE13
+	/* This function tries to increment the counters that limit
+	 * the number of simultaenous requests being processed. It
+	 * assumes that the mutex is held when the function is called
+	 * and returns with the mutex still held, although it may
+	 * unlock and lock the mutex itself.
+	 */
+
 	apr_status_t rv;
 	
 	int max_active_requests = atoi(conf->max_active_requests);
@@ -1913,21 +2071,20 @@ increment_active_requests (xsp_data *conf)
 	/* Limit the number of concurrent requests. If no
 	 * limiting is in effect (or can't be done because
 	 * there is no dashboard), return the OK status. 
-	 * Same test as in the decrement function. */
-	if (max_active_requests == 0 || !conf->dashboard_mutex || !conf->dashboard)
+	 * Same test as in the decrement function and the
+	 * control panel. */
+	if (!conf->dashboard_mutex || !conf->dashboard)
 		return 1;
 		
-	rv = apr_global_mutex_lock (conf->dashboard_mutex);
-	/* Drop the request on failure to lock the dashboard because
-	 * we don't want to decrement the counter later since we couldn't
-	 * increment it here. */
-	if (rv != APR_SUCCESS)
-		return 0;
-		
 	// From here on, rv holds onto whether we still have
 	// the lock acquired, just in case some error ocurrs
 	// acquiring it during the loop.
-	if (conf->dashboard->active_requests >= max_active_requests) {
+	rv = APR_SUCCESS;
+	
+	/* If any limiting is in effect, and if there are the maximum
+	 * allowed concurrent requests, then we have to hold the request
+	 * for a bit of time. */
+	if (max_active_requests > 0 && conf->dashboard->active_requests >= max_active_requests) {
 		/* We need to wait until the active requests
 		 * go below the maximum. */
 		 
@@ -1937,7 +2094,6 @@ increment_active_requests (xsp_data *conf)
 		 * is max_active_req+max_waiting_req.
 		 */
 		if (conf->dashboard->waiting_requests >= max_waiting_requests) {
-			apr_global_mutex_unlock (conf->dashboard_mutex);
 			ap_log_error (APLOG_MARK, APLOG_ERR, STATUS_AND_SERVER,
 			      "Maximum number of concurrent mod_mono requests to %s reached (%d active, %d waiting). Request dropped.",
 		    	  conf->dashboard_lock_file, max_active_requests, max_waiting_requests);
@@ -1972,9 +2128,8 @@ increment_active_requests (xsp_data *conf)
 	 	// many requests are going, stop processing the
 	 	// request.
 		if (rv == APR_SUCCESS && conf->dashboard->active_requests >= max_active_requests) {
-			apr_global_mutex_unlock (conf->dashboard_mutex);
 			ap_log_error (APLOG_MARK, APLOG_ERR, STATUS_AND_SERVER,
-				      "Maximum number (%d) of concurrent mod_mono requests to %s reached. Droping request.",
+				      "Maximum number (%d) of concurrent mod_mono requests to %s reached. Dropping request.",
 				      max_active_requests, conf->dashboard_lock_file);
 			return 0;
 		}
@@ -1987,21 +2142,22 @@ increment_active_requests (xsp_data *conf)
 		return 0;
 
 	conf->dashboard->active_requests++;
-	apr_global_mutex_unlock (conf->dashboard_mutex);
 	
 	return 1;
+#else
+	return 1;
+#endif
 }
 
-static void
+ static void
 decrement_active_requests (xsp_data *conf)
 {
+#ifndef APACHE13
 	apr_status_t rv;
 
-	int max_active_requests = atoi(conf->max_active_requests);
-
 	/* Check if limiting is in effect. Same test as in the
-	 * increment function. */
-	if (max_active_requests == 0 || !conf->dashboard_mutex || !conf->dashboard)
+	 * increment function and the control panel. */
+	if (!conf->dashboard_mutex || !conf->dashboard)
 		return;
 		
 	rv = apr_global_mutex_lock (conf->dashboard_mutex);
@@ -2012,6 +2168,7 @@ decrement_active_requests (xsp_data *conf)
 
 	if (rv == APR_SUCCESS)
 		apr_global_mutex_unlock (conf->dashboard_mutex);
+#endif
 }
 
 static int
@@ -2033,13 +2190,15 @@ mono_execute_request (request_rec *r, char auto_app)
 	int connect_attempts;
 	int start_wait_time;
 	char *socket_name = NULL;
-  
+	int retrying, was_starting;
+	int is_global = 0;
+
 	config = ap_get_module_config (r->server->module_config, &mono_module);
-	DEBUG_PRINT (2, "config = 0x%lx", (uint64_t) config);
+	DEBUG_PRINT (2, "config = 0x%p", config);
 	if (r->per_dir_config != NULL)
 		dir_config = ap_get_module_config (r->per_dir_config, &mono_module);
 
-	DEBUG_PRINT (2, "dir_config = 0x%lx", (uint64_t) dir_config);
+	DEBUG_PRINT (2, "dir_config = 0x%p", dir_config);
 	if (dir_config != NULL && dir_config->alias != NULL)
 		idx = search_for_alias (dir_config->alias, config);
 	else
@@ -2049,8 +2208,10 @@ mono_execute_request (request_rec *r, char auto_app)
 	
 	if (idx < 0) {
 		DEBUG_PRINT (2, "Alias not found. Checking for auto-applications.");
-		if (config->auto_app)
+		if (config->auto_app) {
 			idx = search_for_alias ("XXGLOBAL", config);
+			is_global = 1;
+		}
 
 		if (idx == -1) {
 			DEBUG_PRINT (2, "Global config not found. Finishing request.");
@@ -2098,12 +2259,31 @@ mono_execute_request (request_rec *r, char auto_app)
 		} else
 			conf->dashboard_mutex_initialized_in_child = 1;
 	}
+	
+	if (conf->dashboard_mutex && conf->dashboard) {
+		if (apr_global_mutex_lock (conf->dashboard_mutex) == APR_SUCCESS) {
+			int ok = conf->dashboard->accepting_requests;
+			if (ok)
+				ok = increment_active_requests (conf);
+			
+			apr_global_mutex_unlock (conf->dashboard_mutex);
+
+			if (!ok)
+				return HTTP_SERVICE_UNAVAILABLE;
+		}
+	}
 #endif
-	
+	if (conf->portability_level > PORTABILITY_MAX)
+		conf->portability_level = PORTABILITY_UNKNOWN;
+
+	mono_portability_helpers_init (&conf->portability_level, conf->iomap);
+
 	rv = -1; /* avoid a warning about uninitialized value */
-	if (!increment_active_requests (conf))
-		return HTTP_SERVICE_UNAVAILABLE;
-	
+
+#ifndef APACHE13
+	retrying = connect_attempts;
+	was_starting = 0;
+#endif
 	while (connect_attempts--) {
 		rv = setup_socket (&sock, conf, r->pool);
 		DEBUG_PRINT (2, "After setup_socket");
@@ -2129,12 +2309,37 @@ mono_execute_request (request_rec *r, char auto_app)
 				decrement_active_requests (conf);
 				return HTTP_SERVICE_UNAVAILABLE;
 			}
+
+			if (conf->dashboard->starting) {
+				retrying--;
+				was_starting = 1;
+
+				rv2 = apr_global_mutex_unlock (conf->dashboard_mutex);
+				if (rv2 != APR_SUCCESS)
+					ap_log_error (APLOG_MARK, APLOG_ALERT, STATCODE_AND_SERVER (rv2),
+						      "Failed to release %s lock, the process may deadlock!",
+						      conf->dashboard_lock_file);
+				if (retrying < 0) {
+					ap_log_error (APLOG_MARK, APLOG_CRIT, STATUS_AND_SERVER,
+						      "Another process is attempting to start the backend. This request will fail.");
+					decrement_active_requests (conf);
+					return HTTP_SERVICE_UNAVAILABLE;
+				}
+				connect_attempts++; /* keep trying */
+				apr_sleep (apr_time_from_sec (start_wait_time));
+				continue;
+			}
 #endif
-			
-			if (socket_name != NULL && unlink (socket_name) < 0 && errno != ENOENT)
-				ap_log_error (APLOG_MARK, APLOG_ALERT, STATUS_AND_SERVER,
-					      "Could not remove stale socket %s. %s. Further requests will probably fail.",
-					      socket_name, strerror (errno));
+
+			if (was_starting) {
+				was_starting = 0;
+				connect_attempts++;
+
+				/* let's try one more time here */
+				apr_sleep (apr_time_from_sec (start_wait_time));
+				continue;
+			}
+
 			start_xsp (config, 0, conf->alias);
 			/* give some time for warm-up */
 			DEBUG_PRINT (2, "Started new backend, sleeping %us to let it configure", (unsigned)start_wait_time);
@@ -2369,6 +2574,9 @@ start_xsp (module_cfg *config, int is_restart, char *alias)
 
 #ifdef APACHE13
 		sock = apr_pcalloc (pconf, sizeof (apr_socket_t));
+#else
+		if (xsp->dashboard)
+			xsp->dashboard->starting = 1;
 #endif
 		rv = setup_socket (&sock, xsp, pconf);
 
@@ -2399,6 +2607,10 @@ start_xsp (module_cfg *config, int is_restart, char *alias)
 			}
 #endif
 		}
+#ifndef APACHE13
+		if (xsp->dashboard)
+			xsp->dashboard->starting = 0;
+#endif
 	}
 }
 
@@ -2507,6 +2719,31 @@ terminate_xsp (void *data)
 	return terminate_xsp2(data, NULL, 0, 0);
 }
 
+static void
+set_accepting_requests (void *data, char *alias, int accepting_requests)
+{
+	server_rec *server;
+	module_cfg *config;
+	xsp_data *xsp;
+	int i;
+	
+	server = (server_rec *) data;
+	config = ap_get_module_config (server->module_config, &mono_module);
+
+	for (i = 0; i < config->nservers; i++) {
+		xsp = &config->servers [i];
+
+		/* If alias isn't null, skip XSPs that don't have that alias. */
+		if (alias != NULL && strcmp(xsp->alias, alias))
+			continue;
+		
+#ifndef APACHE13
+		if (xsp->dashboard)
+			xsp->dashboard->accepting_requests = accepting_requests;
+#endif
+	}
+}
+
 static int
 mono_control_panel_handler (request_rec *r)
 {
@@ -2534,43 +2771,63 @@ mono_control_panel_handler (request_rec *r)
 	uri = &r->parsed_uri;
 	if (!uri->query || !strcmp (uri->query, "")) {
 		/* No query string -> Emit links for configuration commands. */
-		request_send_response_string (r, "<ul style=\"text-align: center;\">\n");
+		request_send_response_string (r, "<ul>\n");
+		request_send_response_string (r, "<li><div>All Backends</div>\n<ul>\n");
 		request_send_response_string (r, "<li><a href=\"?restart=ALL\">Restart all mod-mono-server processes</a></li>\n");
+		request_send_response_string (r, "<li><a href=\"?pause=ALL\">Stop Accepting Requests</a></li>\n");
+		request_send_response_string (r, "<li><a href=\"?resume=ALL\">Resume Accepting Requests</a></li>\n");
+		request_send_response_string (r, "</ul></li>\n");
 
 		for (i = 0; i < config->nservers; i++) {
 			xsp = &config->servers [i];
 			if (xsp->run_xsp && !strcasecmp (xsp->run_xsp, "false"))
 				continue;
 
-			buffer = apr_psprintf (r->pool, "<li>%s: <a href=\"?restart=%s\">"
-					       "Restart Server</a>", xsp->alias, xsp->alias);
+			buffer = apr_psprintf (r->pool, "<li><div>%s</div><ul>\n", xsp->alias);
+			request_send_response_string(r, buffer);
+
+			buffer = apr_psprintf (r->pool, "<li><a href=\"?restart=%s\">Restart Server</a></li>\n", xsp->alias);
 			request_send_response_string(r, buffer);
 
 #ifndef APACHE13
 			ensure_dashboard_initialized (config, xsp, pconf);
-			if (xsp->dashboard_mutex) {
-				rv = apr_global_mutex_lock (xsp->dashboard_mutex);
-				if (rv == APR_SUCCESS) {
-					if (xsp->restart_mode == AUTORESTART_MODE_REQUESTS) {
-						buffer = apr_psprintf (r->pool, " [%d requests served; limit: %d]",
-								       xsp->dashboard->handled_requests, xsp->restart_requests);
-						request_send_response_string(r, buffer);
-					} else if (xsp->restart_mode == AUTORESTART_MODE_TIME) {
-						buffer = apr_psprintf (r->pool, " [%ds time running; limit: %ds]",
-								       (int)(time(NULL) - xsp->dashboard->start_time), (int)xsp->restart_time);
-						request_send_response_string(r, buffer);
-					}
+			if (xsp->dashboard_mutex && xsp->dashboard
+				&& apr_global_mutex_lock (xsp->dashboard_mutex) == APR_SUCCESS) {
 
-					rv = apr_global_mutex_unlock (xsp->dashboard_mutex);
-					if (rv != APR_SUCCESS)
-						ap_log_error (APLOG_MARK, APLOG_ALERT, STATCODE_AND_SERVER (rv),
-							      "Failed to release %s lock after mono-ctrl output, the process may deadlock!",
-							      xsp->dashboard_lock_file);
+				if (xsp->dashboard->accepting_requests)
+					buffer = apr_psprintf (r->pool, "<li><a href=\"?pause=%s\">Stop Accepting Requests</a></li>\n", xsp->alias);
+				else
+					buffer = apr_psprintf (r->pool, "<li><a href=\"?resume=%s\">Resume Accepting Requests</a></li>\n", xsp->alias);
+				request_send_response_string(r, buffer);
+				
+				if (xsp->restart_mode == AUTORESTART_MODE_REQUESTS) {
+					buffer = apr_psprintf (r->pool, "<li>%d requests served; limit: %d</li>\n",
+							       xsp->dashboard->handled_requests, xsp->restart_requests);
+					request_send_response_string(r, buffer);
+				} else if (xsp->restart_mode == AUTORESTART_MODE_TIME) {
+					buffer = apr_psprintf (r->pool, "<li>%ds time running; limit: %ds</li>\n",
+							       (int)(time(NULL) - xsp->dashboard->start_time), (int)xsp->restart_time);
+					request_send_response_string(r, buffer);
 				}
+
+				buffer = apr_psprintf (r->pool, "<li>%d requests currently being processed; limit: %s</li>\n",
+						       xsp->dashboard->active_requests, xsp->max_active_requests);
+				request_send_response_string(r, buffer);
+				
+				buffer = apr_psprintf (r->pool, "<li>%d requests currently waiting to be processed; limit: %s</li>\n",
+						       xsp->dashboard->waiting_requests, xsp->max_waiting_requests);
+				request_send_response_string(r, buffer);
+				
+
+				rv = apr_global_mutex_unlock (xsp->dashboard_mutex);
+				if (rv != APR_SUCCESS)
+					ap_log_error (APLOG_MARK, APLOG_ALERT, STATCODE_AND_SERVER (rv),
+						      "Failed to release %s lock after mono-ctrl output, the process may deadlock!",
+						      xsp->dashboard_lock_file);
 			}
 #endif
 			
-			request_send_response_string(r, "</li>\n");
+			request_send_response_string(r, "</ul></li>\n");
 		}
 		
 		request_send_response_string (r, "</ul>\n");
@@ -2580,9 +2837,25 @@ mono_control_panel_handler (request_rec *r)
 			char *alias = uri->query + 8; /* +8 == .Substring(8) */
 			if (!strcmp (alias, "ALL"))
 				alias = NULL;
+			set_accepting_requests (r->server, alias, 0);
 			terminate_xsp2 (r->server, alias, 1, 0); 
 			start_xsp (config, 1, alias);
+			set_accepting_requests (r->server, alias, 1);
 			request_send_response_string (r, "<div style=\"text-align: center;\">mod-mono-server processes restarted.</div><br>\n");
+		} else if (uri->query && !strncmp (uri->query, "pause=", 6)) {
+			/* Stop accepting requests for this server */
+			char *alias = uri->query + 6; /* +6 == .Substring(6) */
+			if (!strcmp (alias, "ALL"))
+				alias = NULL;
+			set_accepting_requests (r->server, alias, 0);
+			request_send_response_string (r, "<div style=\"text-align: center;\">no longer accepting requests</div><br>\n");
+		} else if (uri->query && !strncmp (uri->query, "resume=", 7)) {
+			/* Resume accepting requests for this server */
+			char *alias = uri->query + 7; /* +7 == .Substring(7) */
+			if (!strcmp (alias, "ALL"))
+				alias = NULL;
+			set_accepting_requests (r->server, alias, 1);
+			request_send_response_string (r, "<div style=\"text-align: center;\">resumed accepting requests</div><br>\n");
 		} else {
 			/* Invalid command. */
 			request_send_response_string (r, "<div style=\"text-align: center;\">Invalid query string command.</div>\n");
@@ -2821,6 +3094,10 @@ static const command_rec mono_cmds [] = {
 		    " Default value: Default: \"\""
 	),
 
+	MAKE_CMD12 (MonoIOMAP, iomap,
+		    "A string with format the same as the MONO_IOMAP variable (see mod_mono (1))."
+		    " Default value: none"),
+
 	MAKE_CMD_ITERATE2 (AddMonoApplications, applications,
 			   "Appends an application."
 	),
